@@ -7,6 +7,7 @@ from bamt.redef_HC import hc as hc_method
 from bamt import Nodes
 from bamt.log import logger_builder
 from pandas import DataFrame
+from bamt.utils import GraphUtils as gru
 
 from typing import Dict, List, Optional, Tuple, Callable, TypedDict, Union
 
@@ -51,9 +52,9 @@ class StructureBuilder(object):
         node_type = self.descriptor['types']
         blacklist = []
         datacol = data.columns.to_list()
-
+        
         if not self.has_logit:
-
+          
             # Has_logit flag allows BN building edges between cont and disc
             RESTRICTIONS = [('cont', 'disc'), ('cont', 'disc_num')]
             for x, y in itertools.product(datacol, repeat=2):
@@ -102,6 +103,11 @@ class StructureBuilder(object):
             self.skeleton['V'][id].cont_parents = cont_parents
             self.skeleton['V'][id].children = children
 
+        ordered = gru.toporder(self.skeleton['V'], self.skeleton['E'])
+        notOrdered = [node.name for node in self.skeleton['V']]
+        mask = [notOrdered.index(name) for name in ordered]
+        self.skeleton['V'] = [self.skeleton['V'][i] for i in mask]
+
 
 class VerticesDefiner(StructureBuilder):
     """
@@ -196,7 +202,7 @@ class HillClimbDefiner(EdgesDefiner, VerticesDefiner):
         self.optimizer = HillClimbSearch(data)
         self.params = {'init_edges': None,
                        'init_nodes': None,
-                       'remove_init_edges': False,
+                       'remove_init_edges': True,
                        'white_list': None,
                        'bl_add': None}
         super(HillClimbDefiner, self).__init__(descriptor)
@@ -213,7 +219,7 @@ class HillClimbDefiner(EdgesDefiner, VerticesDefiner):
             logger_builder.error(
                 f"K2 deals only with discrete data. Continuous data: {[col for col, type in BasePreprocessor.get_nodes_types(data).items() if type not in ['disc', 'disc_num']]}")
             return None
-
+        
         if len(self.scoring_function) != 2:
             from pgmpy.estimators import K2Score
             scoring_function = K2Score
@@ -227,9 +233,11 @@ class HillClimbDefiner(EdgesDefiner, VerticesDefiner):
                 white_list=white_list
             )
         else:
+            
             if remove_init_edges:
                 startdag = DAG()
-                startdag.add_nodes_from(nodes=self.vertices)
+                nodes = [str(v) for v in self.vertices]
+                startdag.add_nodes_from(nodes=nodes)
                 startdag.add_edges_from(ebunch=init_edges)
                 best_model = self.optimizer.estimate(black_list=self.black_list, white_list=white_list,
                                                      start_dag=startdag, show_progress=False)
@@ -282,8 +290,8 @@ class HCStructureBuilder(HillClimbDefiner):
         :param data: train data
         :param descriptor: map for data
         """
-
-
+       
+        
         super(HCStructureBuilder, self).__init__(descriptor=descriptor, data=data,
                                                  scoring_function=scoring_function)
         self.use_mixture = use_mixture
@@ -298,6 +306,8 @@ class HCStructureBuilder(HillClimbDefiner):
 
         init_nodes = self.params.pop('init_nodes')
         bl_add = self.params.pop('bl_add')
+   
+
 
         # Level 1
         self.skeleton['V'] = self.vertices
@@ -309,7 +319,7 @@ class HCStructureBuilder(HillClimbDefiner):
             self.apply_group1(data=data, **self.params)
 
         # Level 2
-
+      
         self.get_family()
         self.overwrite_vertex(has_logit=self.has_logit,
                               use_mixture=self.use_mixture,
