@@ -1,5 +1,3 @@
-import bamt.utils.GraphUtils
-from bamt import Builders, Nodes
 import random
 import re
 import networkx as nx
@@ -10,16 +8,18 @@ import numpy as np
 import json
 import os
 
+# from sklearn import preprocessing as pp
 from tqdm import tqdm
 from concurrent.futures import ThreadPoolExecutor
-from bamt.utils import GraphUtils as gru
 from pyvis.network import Network
-
+from pyitlib import discrete_random_variable as drv
 from typing import Dict, Tuple, List, Callable, Optional, Type, Union, Any, Sequence
-from bamt.Builders import ParamDict
 
+from bamt.Builders import ParamDict
 from bamt.log import logger_network
 from bamt.config import config
+from bamt import Builders, Nodes
+# from bamt.Preprocessors import Preprocessor
 
 STORAGE = config.get('NODES', 'models_storage', fallback='models_storage is not defined')
 
@@ -40,6 +40,7 @@ class BaseNetwork(object):
         self._allowed_dtypes = ['Abstract']
         self.nodes = []
         self.edges = []
+        self.weights = {}
         self.descriptor = {"types": {},
                            "signs": {}}
         self.distributions = {}
@@ -113,7 +114,6 @@ class BaseNetwork(object):
                                                  scoring_function=scoring_function,
                                                  has_logit=self.has_logit,
                                                  use_mixture=self.use_mixture)
-       
            
             self.sf_name = scoring_function[0]
             
@@ -123,13 +123,51 @@ class BaseNetwork(object):
             self.nodes = worker.skeleton['V']
             self.edges = worker.skeleton['E']
 
+    def calculate_weights(self, discretized_data: pd.DataFrame):
+        """
+        Provide calculation of link strength according mutual information between node and its parent(-s) values.
+        """
+        import bamt.utils.GraphUtils as gru
+        if not all([i in ['disc', 'disc_num'] for i in gru.nodes_types(discretized_data).values()]):
+            logger_network.error(f"calculate_weghts() method deals only with discrete data. Continuous data: " +
+                                 f"{[col for col, type in gru.nodes_types(discretized_data).items() if type not in ['disc', 'disc_num']]}")
+        if not self.edges:
+            logger_network.error("Bayesian Network hasn't fitted yet. Please add edges with add_edges() method")
+        if not self.nodes:
+            logger_network.error("Bayesian Network hasn't fitted yet. Please add nodes with add_nodes() method")
+        weights = dict()
+
+        for node in self.nodes:
+            parents = node.cont_parents + node.disc_parents
+            if parents is None:
+                continue
+            y = discretized_data[node.name].values
+            if len(parents) == 1:
+                x = discretized_data[parents[0]].values
+                LS_true = drv.information_mutual(X=y, Y=x)
+                entropy = drv.entropy(X=y)
+                weight = LS_true / entropy
+                weights[(parents[0], node.name)] = weight
+            else:
+                for parent_node in parents:
+                    x = discretized_data[parent_node].values
+                    other_parents = [tmp for tmp in parents if tmp != parent_node]
+                    z = list()
+                    for other_parent in other_parents:
+                        z.append(list(discretized_data[other_parent].values))
+                    LS_true = np.average(drv.information_mutual_conditional(X=y, Y=x, Z=z, cartesian_product=True))
+                    entropy = np.average(drv.entropy_conditional(X=y, Y=z, cartesian_product=True)) + 1e-8
+                    weight = LS_true / entropy
+                    weights[(parent_node, node.name)] = weight
+        self.weights = weights
+
     def set_nodes(self, nodes: List, info: Optional[Dict] = None):
         """
         additional function to set nodes manually. User should be aware that
         nodes must be a subclass of BaseNode.
         :param nodes dict with name and node (if a lot of nodes should be added)
         """
-        if not info and not (self.descriptor["types"] and self.descriptor["signs"]):
+        if not info and not self.descriptor["types"]:
             logger_network.error("In case of manual setting nodes user should set map for them as well.")
             return
         self.nodes = []
@@ -191,7 +229,7 @@ class BaseNetwork(object):
                 builder.get_family()
                 if self.edges:
                     builder.overwrite_vertex(has_logit=self.has_logit, use_mixture=self.use_mixture)
-                    self.set_nodes(builder.vertices)
+                    self.set_nodes(nodes=builder.skeleton['V'])
                 else:
                     logger_network.error("Empty set of edges")
 
@@ -238,6 +276,44 @@ class BaseNetwork(object):
             json.dump(self.edges, out)
         return True
 
+    def save(self, outdir: str):
+        """
+        Function to save the whole BN to json file
+        :param outdir: output directory
+        """
+        if not outdir.endswith('.json'):
+            return None
+        new_weights = dict()
+        for key in self.weights:
+            new_weights[str(key)] = self.weights[key]
+        outdict = {
+            'info': self.descriptor,
+            'edges': self.edges,
+            'parameters': self.distributions,
+            'weights': new_weights
+        }
+        with open(outdir, 'w+') as out:
+            json.dump(outdict, out)
+        return True
+
+    def load(self, input_dir: str):
+        """
+        Function to load the whole BN from json file
+        :param input_dir: input directory
+        """
+        with open(input_dir) as f:
+            input_dict = json.load(f)
+
+        self.add_nodes(input_dict['info'])
+        self.set_structure(edges=input_dict['edges'])
+        self.set_parameters(parameters=input_dict['parameters'])
+        str_keys = list(input_dict['weights'].keys())
+        tuple_keys = [eval(key) for key in str_keys]
+        weights = {}
+        for tuple_key in tuple_keys:
+            weights[tuple_key] = input_dict['weights'][str(tuple_key)]
+        self.weights = weights
+
     def fit_parameters(self, data: pd.DataFrame, dropna: bool = True):
         """
         Base function for parameters learning
@@ -261,12 +337,6 @@ class BaseNetwork(object):
             columns_names = [name for name, t in self.descriptor['types'].items() if t in ['disc_num']]
             data[columns_names] = data.loc[:, columns_names].astype('str')
 
-        # Topology sorting
-        ordered = gru.toporder(self.nodes, self.edges)
-        notOrdered = [node.name for node in self.nodes]
-        mask = [notOrdered.index(name) for name in ordered]
-        self.nodes = [self.nodes[i] for i in mask]
-
         def worker(node):
             return node.fit_parameters(data)
 
@@ -297,15 +367,21 @@ class BaseNetwork(object):
                 print(
                     f"{n.name: <20} | {n.type: <50} | {self.descriptor['types'][n.name]: <10} | {str([self.descriptor['types'][name] for name in n.cont_parents + n.disc_parents]): <50} | {str([name for name in n.cont_parents + n.disc_parents])}")
 
-    def sample(self, n: int, evidence: Optional[Dict[str, Union[str, int, float]]] = None,
-               as_df: bool = True, predict: bool = False) -> Union[
-        None, pd.DataFrame, List[Dict[str, Union[str, int, float]]]]:
+    def sample(self,
+               n: int,
+               evidence: Optional[Dict[str, Union[str, int, float]]] = None,
+               as_df: bool = True,
+               predict: bool = False,
+               parall_count: int = 1) -> \
+            Union[None, pd.DataFrame, List[Dict[str, Union[str, int, float]]]]:
         """
         Sampling from Bayesian Network
         n: int number of samples
         evidence: values for nodes from user
+        parall_count: number of threads. Defaults to 1.
         """
-        seq = []
+        from joblib import Parallel, delayed
+
         random.seed()
         if not self.distributions.items():
             logger_network.error("Parameter learning wasn't done. Call fit_parameters method")
@@ -316,26 +392,12 @@ class BaseNetwork(object):
                     if not (isinstance(evidence[node.name], str)):
                         evidence[node.name] = str(int(evidence[node.name]))
 
-        for n in range(n):
+        def wrapper():
             output = {}
             for node in self.nodes:
                 parents = node.cont_parents + node.disc_parents
-                if evidence:
-                    if node.name in evidence.keys():
-                        output[node.name] = evidence[node.name]
-                    else:
-                        if not parents:
-                            pvals = None
-                        else:
-                            if self.type == 'Discrete':
-                                pvals = [str(output[t]) for t in parents]
-                            else:
-                                pvals = [output[t] for t in parents]
-                        if predict:
-                            output[node.name] = node.predict(self.distributions[node.name], pvals=pvals)
-                        else:
-                            output[node.name] = node.choose(self.distributions[node.name], pvals=pvals)
-
+                if evidence and node.name in evidence.keys():
+                    output[node.name] = evidence[node.name]
                 else:
                     if not parents:
                         pvals = None
@@ -345,11 +407,26 @@ class BaseNetwork(object):
                         else:
                             pvals = [output[t] for t in parents]
                     if predict:
-                        output[node.name] = node.predict(self.distributions[node.name], pvals=pvals)
+                        output[node.name] = \
+                            node.predict(
+                                self.distributions[node.name], pvals=pvals)
                     else:
-                        output[node.name] = node.choose(self.distributions[node.name], pvals=pvals)
+                        output[node.name] = \
+                            node.choose(
+                                self.distributions[node.name], pvals=pvals)
+            return output
 
-            seq.append(output)
+        seq = Parallel(n_jobs=parall_count)(
+            delayed(wrapper)()
+            for i in tqdm(range(n), position=0, leave=True))
+        seq_df = pd.DataFrame.from_dict(seq, orient='columns')
+        seq_df.dropna(inplace=True)
+        cont_nodes = [c.name for c in self.nodes if c.type != 'Discrete' and 'Logit' not in c.type]
+        positive_columns = [c for c in cont_nodes if self.descriptor['signs'][c] == 'pos']
+        seq_df = seq_df[(seq_df[positive_columns] >= 0).all(axis=1)]
+        seq_df.reset_index(inplace=True, drop=True)
+        seq = seq_df.to_dict('records')
+        
 
         if as_df:
             return pd.DataFrame.from_dict(seq, orient='columns')
@@ -369,6 +446,10 @@ class BaseNetwork(object):
         Returns:
             predicted data (dict): dict with column as key and predicted data as value
         """
+        if test.isnull().any().any():
+            logger_network.error("Test data contains NaN values.")
+            return {}
+
         from joblib import Parallel, delayed
 
         def wrapper(bn: HybridBN, test: pd.DataFrame, columns: List[str]):
