@@ -4,9 +4,10 @@ from concurrent.futures import ThreadPoolExecutor
 from pomegranate import DiscreteDistribution, ConditionalProbabilityTable
 from bamt.utils.MathUtils import *
 from gmr import GMM
+# from sklearn.mixture import GaussianMixture
 from pandas import DataFrame
 
-from typing import Dict, List, Any, Union, TypedDict, Type, Optional
+from typing import Dict, List, Any, Union, TypedDict, Type, Optional, Sequence
 
 import itertools
 import random
@@ -131,11 +132,54 @@ class DiscreteNode(BaseNode):
 
         return vals[rindex]
 
+    @staticmethod
+    def predict(node_info: Dict[str, Union[float, str]], pvals: List[str]) -> str:
+        """function for prediction based on evidence values in discrete node
+
+        Args:
+            node_info (Dict[str, Union[float, str]]): parameters of node
+            pvals (List[str]): values in parents nodes
+
+        Returns:
+            str: prediction
+        """        
+        # vals = node_info['vals']
+        # if not pvals:
+        #     dist = node_info['cprob']
+        # else:
+        #     # noinspection PyTypeChecker
+        #     dist = node_info['cprob'][str(pvals)]
+        # index_max = 0    
+        # g = itertools.groupby(dist)
+        # if next(g, True) and not next(g, False):
+        #     index_max = random.randint(0, len(dist))
+        # else:
+        #     index_max = np.argmax(dist)
+        # return vals[index_max]
+
+        vals = node_info['vals']
+        disct = []
+        if not pvals:
+            dist = node_info['cprob']
+        else:
+            # noinspection PyTypeChecker
+            dist = node_info['cprob'][str(pvals)]
+        max_value = max(dist)
+        indices = [index for index, value in enumerate(dist) if value == max_value]
+        max_ind = 0
+        if len(indices) == 1:
+            max_ind = indices[0]
+        else:
+            max_ind = random.choice(indices)
+        return vals[max_ind]
+        
+
+
 
 class GaussianParams(TypedDict):
-    mean: float
-    coef: List[float]
-    variance: float
+    mean: np.ndarray
+    coef: Sequence[Any]
+    variance: Union[np.ndarray, float]
 
 
 class GaussianNode(BaseNode):
@@ -186,12 +230,30 @@ class GaussianNode(BaseNode):
         """
         mean = node_info["mean"]
         if pvals:
-            for m in pvals:
-                mean += m * node_info['coef'][0]
+            for i, m in enumerate(pvals):
+                mean += m * node_info['coef'][i]
         variance = node_info['variance']
         # distribution = [mean, variance]
         return random.gauss(mean, math.sqrt(variance))
 
+    @staticmethod
+    def predict(node_info: Dict[str, Union[float, List[float]]],
+                pvals: List[float]) -> float:
+        """function for prediction in gaussian node
+
+        Args:
+            node_info (Dict[str, Union[float, List[float]]]): node parameters
+            pvals (List[float]): parent values
+
+        Returns:
+            float: prediction
+        """
+        mean = node_info["mean"]
+        if pvals:
+            for i, m in enumerate(pvals):
+                mean += m * node_info['coef'][i]
+        return mean
+
 
 class CondGaussParams(TypedDict):
     variance: Optional[float]
@@ -278,6 +340,31 @@ class ConditionalGaussianNode(BaseNode):
         variance = lgdistribution["variance"]
         return random.gauss(mean, math.sqrt(variance))
 
+    @staticmethod
+    def predict(node_info: Dict[str, Dict[str, CondGaussParams]], pvals: List[Union[str, float]]) -> float:
+        """function for prediction in conditional gaussian node
+
+        Args:
+            node_info (Dict[str, Union[float, List[float]]]): node parameters
+            pvals (List[float]): parent values
+
+        Returns:
+            float: prediction
+        """
+        dispvals = []
+        lgpvals = []
+        for pval in pvals:
+            if (isinstance(pval, str)) | (isinstance(pval, int)):
+                dispvals.append(pval)
+            else:
+                lgpvals.append(pval)
+        lgdistribution = node_info["hybcprob"][str(dispvals)]
+        mean = lgdistribution["mean"]
+        if lgpvals:
+            for x in range(len(lgpvals)):
+                mean += lgpvals[x] * lgdistribution["coef"][x]
+        return mean
+
 
 class MixtureGaussianParams(TypedDict):
     mean: List[float]
@@ -303,8 +390,8 @@ class MixtureGaussianNode(BaseNode):
             n_comp = int((component(data, [self.name], 'aic') + component(data, [self.name],
                                                                           'bic')) / 2)  # component(data, [node], 'LRTS')#
             # n_comp = 3
-            gmm = GMM(n_components=n_comp)
-            gmm.from_samples(np.transpose([data[self.name].values]))
+            gmm = GMM(n_components=n_comp).from_samples(np.transpose([data[self.name].values]), n_iter=500,
+                                                        init_params='kmeans++')
             means = gmm.means.tolist()
             cov = gmm.covariances.tolist()
             # weigts = np.transpose(gmm.to_responsibilities(np.transpose([data[node].values])))
@@ -320,8 +407,7 @@ class MixtureGaussianNode(BaseNode):
                 n_comp = int((component(new_data, nodes, 'aic') + component(new_data, nodes,
                                                                             'bic')) / 2)  # component(new_data, nodes, 'LRTS')#
                 # n_comp = 3
-                gmm = GMM(n_components=n_comp)
-                gmm.from_samples(new_data[nodes].values)
+                gmm = GMM(n_components=n_comp).from_samples(new_data[nodes].values, n_iter=500, init_params='kmeans++')
                 means = gmm.means.tolist()
                 cov = gmm.covariances.tolist()
                 # weigts = np.transpose(gmm.to_responsibilities(new_data[nodes].values))
@@ -360,6 +446,35 @@ class MixtureGaussianNode(BaseNode):
             sample = np.nan
         return sample
 
+    @staticmethod
+    def predict(node_info: MixtureGaussianParams, pvals: List[Union[str, float]]) -> Optional[float]:
+        """
+        Func to get prediction from current node
+        node_info: nodes info from distributions
+        pvals: parent values
+        Return value from MixtureGaussian node
+        """
+        mean = node_info["mean"]
+        covariance = node_info["covars"]
+        w = node_info["coef"]
+        n_comp = len(node_info['coef'])
+        if n_comp != 0:
+            if pvals:
+                indexes = [i for i in range(1, len(pvals) + 1)]
+                if not np.isnan(np.array(pvals)).all():
+                    gmm = GMM(n_components=n_comp, priors=w, means=mean, covariances=covariance)
+                    sample = gmm.predict(indexes, [pvals])[0][0]
+                else:
+                    sample = np.nan
+            else:
+                # gmm = GMM(n_components=n_comp, priors=w, means=mean, covariances=covariance)
+                sample = 0
+                for ind, wi in enumerate(w):
+                    sample += wi * mean[ind][0]
+        else:
+            sample = np.nan
+        return sample
+
 
 class CondMixtureGaussParams(TypedDict):
     mean: Optional[List[float]]
@@ -402,14 +517,14 @@ class ConditionalMixtureGaussianNode(BaseNode):
                     n_comp = int((component(new_data, nodes, 'aic') + component(new_data, nodes,
                                                                                 'bic')) / 2)  # component(new_data, nodes, 'LRTS')#int((component(new_data, nodes, 'aic') + component(new_data, nodes, 'bic')) / 2)
                     # n_comp = 3
-                    gmm = GMM(n_components=n_comp)
-                    gmm.from_samples(new_data[nodes].values)
+                    gmm = GMM(n_components=n_comp).from_samples(new_data[nodes].values, n_iter=500,
+                                                                init_params='kmeans++')
                 else:
                     n_comp = int((component(new_data, [self.name], 'aic') + component(new_data, [self.name],
                                                                                       'bic')) / 2)  # component(new_data, [node], 'LRTS')#int((component(new_data, [node], 'aic') + component(new_data, [node], 'bic')) / 2)
                     # n_comp = 3
-                    gmm = GMM(n_components=n_comp)
-                    gmm.from_samples(np.transpose([new_data[self.name].values]))
+                    gmm = GMM(n_components=n_comp).from_samples(np.transpose([new_data[self.name].values]), n_iter=500,
+                                                                init_params='kmeans++')
                 means = gmm.means.tolist()
                 cov = gmm.covariances.tolist()
                 # weigts = np.transpose(gmm.to_responsibilities(np.transpose([new_data[node].values])))
@@ -476,6 +591,46 @@ class ConditionalMixtureGaussianNode(BaseNode):
             sample = np.nan
         return sample
 
+    @staticmethod
+    def predict(node_info: Dict[str, Dict[str, CondMixtureGaussParams]],
+                pvals: List[Union[str, float]]) -> Optional[float]:
+        """
+        Function to get prediction from ConditionalMixtureGaussian node
+        params:
+        node_info: nodes info from distributions
+        pvals: parent values
+        """
+
+        dispvals = []
+        lgpvals = []
+        for pval in pvals:
+            if ((isinstance(pval, str)) | ((isinstance(pval, int)))):
+                dispvals.append(pval)
+            else:
+                lgpvals.append(pval)
+        lgdistribution = node_info["hybcprob"][str(dispvals)]
+        mean = lgdistribution["mean"]
+        covariance = lgdistribution["covars"]
+        w = lgdistribution["coef"]
+        if len(w) != 0:
+            if len(lgpvals) != 0:
+                indexes = [i for i in range(1, (len(lgpvals) + 1), 1)]
+                if not np.isnan(np.array(lgpvals)).all():
+                    n_comp = len(w)
+                    gmm = GMM(n_components=n_comp, priors=w, means=mean, covariances=covariance)
+                    sample = gmm.predict(indexes, [lgpvals])[0][0]
+                else:
+                    sample = np.nan
+            else:
+                # n_comp = len(w)
+                # gmm = GMM(n_components=n_comp, priors=w, means=mean, covariances=covariance)
+                sample = 0
+                for ind, wi in enumerate(w):
+                    sample += wi * mean[ind][0]
+        else:
+            sample = np.nan
+        return sample
+
 
 class LogitParams(TypedDict):
     classes: List[int]
@@ -563,6 +718,30 @@ class LogitNode(BaseNode):
         else:
             return str(node_info["classes"][0])
 
+    def predict(self, node_info: LogitParams, pvals: List[Union[str, float]]) -> str:
+        """
+        Return prediction from Logit node
+        params:
+        node_info: nodes info from distributions
+        pvals: parent values
+        """
+        pvals = [str(p) for p in pvals]
+
+        if len(node_info["classes"]) > 1:
+            if node_info["serialization"] == 'joblib':
+                model = joblib.load(node_info["classifier_obj"])
+            else:
+                # str_model = node_info["classifier_obj"].decode('latin1').replace('\'', '\"')
+                a = node_info["classifier_obj"].encode('latin1')
+                model = pickle.loads(a)
+
+            pred = model.predict(np.array(pvals).reshape(1, -1))[0]
+
+            return str(pred)
+
+        else:
+            return str(node_info["classes"][0])
+
 
 class ConditionalLogitNode(BaseNode):
     """
@@ -684,3 +863,37 @@ class ConditionalLogitNode(BaseNode):
 
         else:
             return str(lgdistribution["classes"][0])
+
+    def predict(self, node_info: Dict[str, Dict[str, LogitParams]], pvals: List[Union[str, float]]) -> str:
+        """
+        Return value from ConditionalLogit node
+        params:
+        node_info: nodes info from distributions
+        pvals: parent values
+        """
+
+        dispvals = []
+        lgpvals = []
+        for pval in pvals:
+            if (isinstance(pval, str)):
+                dispvals.append(pval)
+            else:
+                lgpvals.append(pval)
+
+        lgdistribution = node_info["hybcprob"][str(dispvals)]
+
+        # JOBLIB
+        if len(lgdistribution["classes"]) > 1:
+            if lgdistribution["serialization"] == 'joblib':
+                model = joblib.load(lgdistribution["classifier_obj"])
+            else:
+                # str_model = lgdistribution["classifier_obj"].decode('latin1').replace('\'', '\"')
+                bytes_model = lgdistribution["classifier_obj"].encode('latin1')
+                model = pickle.loads(bytes_model)
+
+            pred = model.predict(np.array(lgpvals).reshape(1, -1))[0]
+
+            return str(pred)
+
+        else:
+            return str(lgdistribution["classes"][0])
